@@ -1,11 +1,14 @@
-from typing import List, Optional
+from typing import (
+    List,
+    Optional,
+)
 
+from jax import vmap
 import jax.numpy as jnp
 import numpy as np
-from jax import vmap
 
-from fortuna.typing import Array
 from fortuna.conformal.classification.base import ConformalClassifier
+from fortuna.typing import Array
 
 
 class AdaptivePredictionConformalClassifier(ConformalClassifier):
