@@ -1,10 +1,13 @@
-from typing import List, Optional
+from typing import (
+    List,
+    Optional,
+)
 
-import jax.numpy as jnp
 from jax import vmap
+import jax.numpy as jnp
 
-from fortuna.typing import Array
 from fortuna.conformal.classification.base import ConformalClassifier
+from fortuna.typing import Array
 
 
 class SimplePredictionConformalClassifier(ConformalClassifier):
