@@ -1,4 +1,9 @@
-from typing import Dict, Optional, Tuple, Union
+from typing import (
+    Dict,
+    Optional,
+    Tuple,
+    Union,
+)
 
 import jax.nn
 import jax.numpy as jnp
@@ -84,6 +89,8 @@ def compute_counts_confs_accs(
 
     if plot:
         idx = confs != 0
+        if plot_options is None:
+            plot_options = dict()
         plot_reliability_diagram(accs[idx], confs[idx], **plot_options)
 
     return counts, confs, accs
@@ -204,12 +211,7 @@ def brier_score(probs: Array, targets: Union[TargetsLoader, Array]) -> jnp.ndarr
     jnp.ndarray
         The Brier score.
     """
-    if probs.ndim != 2:
-        raise ValueError(
-            """`probs` must be a two-dimensional array of probabilities for each class and each data
-        point."""
-        )
     if type(targets) == TargetsLoader:
         targets = targets.to_array_targets()
-    targets = jax.nn.one_hot(targets, probs.shape[1])
+    targets = jax.nn.one_hot(targets, probs.shape[-1])
     return jnp.mean(jnp.sum((probs - targets) ** 2, axis=1))
