@@ -3,8 +3,14 @@ from fortuna.model.hyper import HyperparameterModel
 from fortuna.model.lenet import LeNet5
 from fortuna.model.linear import LinearModel
 from fortuna.model.mlp import MLP
-from fortuna.model.resnet import (ResNet18, ResNet34, ResNet50, ResNet101,
-                                  ResNet152, ResNet200)
+from fortuna.model.resnet import (
+    ResNet18,
+    ResNet34,
+    ResNet50,
+    ResNet101,
+    ResNet152,
+    ResNet200,
+)
 from fortuna.model.scalar_constant import ScalarConstantModel
 from fortuna.model.scalar_hyper import ScalarHyperparameterModel
 from fortuna.model.wideresnet import WideResNet28_10
