@@ -1,18 +1,28 @@
 import abc
 import logging
-from typing import Callable, Dict, Optional
+from typing import (
+    Callable,
+    Dict,
+    Optional,
+)
 
 import jax
 import jax.numpy as jnp
 
-from fortuna.calibration.state import CalibState
 from fortuna.data.loader import DataLoader
+from fortuna.output_calib_model.state import OutputCalibState
 from fortuna.prob_model.calib_config.base import CalibConfig
-from fortuna.prob_model.fit_config import FitConfig
+from fortuna.prob_model.fit_config.base import FitConfig
 from fortuna.prob_model.prob_model_calibrator import (
-    JittedProbModelCalibrator, MultiDeviceProbModelCalibrator,
-    ProbModelCalibrator)
-from fortuna.typing import Array, Path, Status
+    JittedProbModelOutputCalibrator,
+    MultiDeviceProbModelOutputCalibrator,
+    ProbModelOutputCalibrator,
+)
+from fortuna.typing import (
+    Array,
+    Path,
+    Status,
+)
 from fortuna.utils.data import check_data_loader_is_not_random
 from fortuna.utils.device import select_trainer_given_devices
 from fortuna.utils.random import RandomNumberGenerator
@@ -85,7 +95,6 @@ class ProbModel(abc.ABC):
             fit_config=fit_config,
             map_fit_config=map_fit_config,
         )
-        logging.info("Fit completed.")
 
         calib_status = None
         if calib_data_loader:
@@ -107,16 +116,19 @@ class ProbModel(abc.ABC):
         check_data_loader_is_not_random(calib_data_loader)
         if val_data_loader is not None:
             check_data_loader_is_not_random(val_data_loader)
-        if self.output_calib_manager is None:
+        if (
+            self.output_calib_manager is None
+            or self.output_calib_manager.output_calibrator is None
+        ):
             logging.warning(
                 """Nothing to calibrate. No calibrator was passed to the probabilistic model."""
             )
         else:
             if self.posterior.state is None:
                 raise ValueError(
-                    """Before calibration, you must either train the probabilistic model (see 
-                        :meth:`~fortuna.prob_model.base.ProbModel.train`), 
-                        or load a state from an existing checkpoint 
+                    """Before calibration, you must either train the probabilistic model (see
+                        :meth:`~fortuna.prob_model.base.ProbModel.train`),
+                        or load a state from an existing checkpoint
                         (see :meth:`~fortuna.prob_model.base.ProbModel.load_state`)."""
                 )
             if calib_config.monitor.verbose:
@@ -124,7 +136,7 @@ class ProbModel(abc.ABC):
                     "Pre-compute ensemble of outputs on the calibration data loader."
                 )
 
-            distribute = jax.local_device_count() > 1
+            distribute = jax.local_devices()[0].platform != "cpu"
 
             (
                 calib_ensemble_outputs_loader,
@@ -152,9 +164,9 @@ class ProbModel(abc.ABC):
 
             trainer_cls = select_trainer_given_devices(
                 devices=calib_config.processor.devices,
-                BaseTrainer=ProbModelCalibrator,
-                JittedTrainer=JittedProbModelCalibrator,
-                MultiDeviceTrainer=MultiDeviceProbModelCalibrator,
+                base_trainer_cls=ProbModelOutputCalibrator,
+                jitted_trainer_cls=JittedProbModelOutputCalibrator,
+                multi_device_trainer_cls=MultiDeviceProbModelOutputCalibrator,
                 disable_jit=calib_config.processor.disable_jit,
             )
 
@@ -176,7 +188,7 @@ class ProbModel(abc.ABC):
             if calib_config.checkpointer.restore_checkpoint_path is None:
                 calib_dict = self.posterior.state.extract_calib_keys()
 
-                state = CalibState.init(
+                state = OutputCalibState.init(
                     params=calib_dict["calib_params"],
                     mutable=calib_dict["calib_mutable"],
                     optimizer=calib_config.optimizer.method,
@@ -192,7 +204,7 @@ class ProbModel(abc.ABC):
             state, status = calibrator.train(
                 rng=self.rng.get(),
                 state=state,
-                fun=self.predictive._batched_log_joint_prob,
+                loss_fun=self.predictive._batched_negative_log_joint_prob,
                 training_data_loader=calib_data_loader,
                 training_dataset_size=calib_size,
                 n_epochs=calib_config.optimizer.n_epochs,
