@@ -1,20 +1,40 @@
 from __future__ import annotations
 
 import abc
-from typing import Any, Callable, Dict, Optional, Tuple, Union
+from typing import (
+    Any,
+    Callable,
+    Dict,
+    List,
+    Optional,
+    Tuple,
+    Union,
+)
 
-import jax.numpy as jnp
-import numpy as np
 from flax.core import FrozenDict
-from jax import random, vmap
+from jax import (
+    random,
+    vmap,
+)
 from jax._src.prng import PRNGKeyArray
+import jax.numpy as jnp
 from jax.tree_util import tree_map
 from optax._src.base import PyTree
 
 from fortuna.distribution.base import Distribution
 from fortuna.prob_model.posterior.posterior_trainer import PosteriorTrainerABC
 from fortuna.prob_model.posterior.state import PosteriorState
-from fortuna.typing import Array, Batch, CalibMutable, CalibParams, Params
+from fortuna.training.callback import Callback
+from fortuna.typing import (
+    Array,
+    Batch,
+    CalibMutable,
+    CalibParams,
+    Mutable,
+    Params,
+    Path,
+)
+from fortuna.utils.strings import encode_tuple_of_lists_of_strings_to_numpy
 
 
 class NormalizingFlowTrainer(PosteriorTrainerABC):
@@ -22,27 +42,39 @@ class NormalizingFlowTrainer(PosteriorTrainerABC):
     def __str__(self):
         pass
 
-    def __init__(self, *, base: Distribution, architecture: object, **kwargs):
+    def sample_forward(self, rng: jnp.array, params: any, n_samples: int) -> tuple:
         """
-        This is a Normalizing Flow. It is able to transform a `base` distribution into a target one via an invertible
-        `architecture`.
+        Sample from the push-forward distribution.
 
-        :param base: object
-            Base distribution. It must include the following methods:
-                - sample(rng: jnp.array, n_samples: int) -> jnp.array
-                    It draws `n_samples` samples from the distribution.
-                - log_joint_prob(x: jnp.array) -> float
-                    It evaluates the log-probability density function at `x`.
-            Common distribution are already available in `prob.distribution`.
-        :param architecture: object
-            Invertible architecture. It must include the following methods:
-                - forward(params: jnp.array, u: jnp.array) ->  jnp.array
-                    It applies on `u` the transformation parametrized by `params`.
-                - backward(params: jnp.array, v: jnp.array) ->  jnp.array
-                    It applies on `v` the inverse transformation parametrized by `params`.
-                - init_params(rng: jnp.array)
-                    It initializes the parameters of the transformation.
+        Parameters
+        ----------
+        rng: PRNGKeyArray
+            Random number generator.
+        params: Params
+            Transformation parameters.
+        n_samples: int
+            Number of samples to draw.
+
+        Returns
+        -------
+        tuple
+            forward_samples: jnp.ndarray
+                Samples.
+            ldj: jnp.ndarray
+                Log-determinant of the Jacobians.
         """
+        return self.forward(params, self.sample_base(rng, n_samples))
+
+    def __init__(
+        self,
+        *,
+        base: Distribution,
+        architecture: object,
+        which_params: Optional[Tuple[List[str]]],
+        unravel: Callable,
+        sub_unravel: Callable,
+        **kwargs,
+    ):
         super(NormalizingFlowTrainer, self).__init__(**kwargs)
         # base distribution
         self.sample_base = base.sample
@@ -53,32 +85,21 @@ class NormalizingFlowTrainer(PosteriorTrainerABC):
         self.backward = architecture.backward
         self.init_params = architecture.init_params
 
-    def sample_forward(self, rng: jnp.array, params: any, n_samples: int) -> tuple:
-        """
-        Sample from the push-forward distribution.
-
-        :param rng: jnp.array
-            Random number generator.
-        :param params: any
-            Transformation parameters.
-        :param n_samples:
-            Number of samples to draw.
-
-        :return: tuple
-            forward_samples: jnp.array
-                Samples.
-            ldj: jnp.array
-                Log-determinant of the Jacobians.
-        """
-        return self.forward(params, self.sample_base(rng, n_samples))
+        # Normalizing flows on subsets of the model parameters
+        self._which_params = which_params
+        self._encoded_which_params = encode_tuple_of_lists_of_strings_to_numpy(
+            which_params
+        )
+        self._unravel = unravel
+        self._sub_unravel = sub_unravel
 
     def training_loss_step(
         self,
-        fun: Callable[[Any], Union[float, Tuple[float, dict]]],
+        loss_fun: Callable[[Any], Union[float, Tuple[float, dict]]],
         params: Params,
-        batch: Tuple[Union[jnp.ndarray, np.ndarray], Union[jnp.ndarray, np.ndarray]],
-        mutable: FrozenDict[str, FrozenDict],
-        rng: jnp.ndarray,
+        batch: Batch,
+        mutable: Mutable,
+        rng: PRNGKeyArray,
         n_data: int,
         unravel: Optional[Callable[[any], PyTree]] = None,
         calib_params: Optional[CalibParams] = None,
@@ -89,10 +110,10 @@ class NormalizingFlowTrainer(PosteriorTrainerABC):
         if mutable is not None:
             return_aux += ["mutable"]
         rng, key = random.split(rng)
-        v, ldj = self.sample_forward(key, tuple(params.values()), kwargs["n_samples"])
-        logp, aux = vmap(
-            lambda _params: fun(
-                unravel(_params),
+        v, ldj = self.sample_forward(key, params, kwargs["n_samples"])
+        neg_logp, aux = vmap(
+            lambda _rav_params: loss_fun(
+                unravel(_rav_params),
                 batch,
                 n_data=n_data,
                 mutable=mutable,
@@ -104,7 +125,7 @@ class NormalizingFlowTrainer(PosteriorTrainerABC):
             )
         )(v)
         return (
-            -(jnp.mean(logp) + jnp.mean(ldj)),
+            jnp.mean(neg_logp) - jnp.mean(ldj),
             {
                 "outputs": aux.get("outputs"),
                 "mutable": tree_map(lambda x: x[-1], aux.get("mutable")),
@@ -119,33 +140,15 @@ class NormalizingFlowTrainer(PosteriorTrainerABC):
         aux: Dict[str, Any],
         batch: Batch,
         metrics: Optional[Tuple[Callable[[jnp.ndarray, Array], float], ...]] = None,
+        callbacks: Optional[List[Callback]] = None,
         kwargs: FrozenDict[str, Any] = FrozenDict(),
-    ) -> Dict[str, jnp.ndarray]:
-        """
-        Perform training batch metrics_names computation and save checkpoint if needed.
-
-        :param current_epoch: int
-            Current epoch.
-        :param state: TrainState
-            The training state.
-        :param aux: Dict[str, Any]
-            The dictionary obtained from a call to `training_loss_step()`. It must have a key name `outputs` which
-            contains the model's prediction for the given `batch`.
-        :param batch: Batch
-            The input data and the targets.
-        :param metrics: Optional[Tuple[Callable[[jnp.ndarray, Array], float], ...]]
-            A tuple of metrics.
-        :param kwargs: FrozenDict[str, Any]
-            Any other extra argument. They have to be explicitly passed within a dictionary and cannot be provided as
-            named arguments due to `jax.jit`.
-
-        :return: Dict[str, jnp.ndarray]
-            A dictionary containing `metrics_names` values on the given input batch.
-        """
+    ) -> Tuple[PosteriorState, Dict[str, jnp.ndarray]]:
         if (
-            self.save_checkpoint_dir
-            and self.save_every_n_steps
-            and current_epoch % self.save_every_n_steps
+            self.save_checkpoint_dir is not None
+            and self.save_every_n_steps is not None
+            and self.save_every_n_steps > 0
+            and self._global_training_step >= self.save_every_n_steps
+            and self._global_training_step % self.save_every_n_steps == 0
         ):
             self.save_checkpoint(
                 state, self.save_checkpoint_dir, keep=self.keep_top_n_checkpoints
@@ -186,13 +189,14 @@ class NormalizingFlowTrainer(PosteriorTrainerABC):
             )
             for k, v in training_batch_metrics.items():
                 training_losses_and_metrics[k] = v
-        return training_losses_and_metrics
+        state = self._callback_loop(state, callbacks, "training_step_end")
+        return state, training_losses_and_metrics
 
     def validation_step(
         self,
         state: PosteriorState,
         batch: Batch,
-        fun: Callable[[Any], Union[float, Tuple[float, dict]]],
+        loss_fun: Callable[[Any], Union[float, Tuple[float, dict]]],
         rng: PRNGKeyArray,
         n_data: int,
         metrics: Optional[Tuple[Callable[[jnp.ndarray, Array], float]], ...] = None,
@@ -200,12 +204,10 @@ class NormalizingFlowTrainer(PosteriorTrainerABC):
         kwargs: FrozenDict[str, Any] = FrozenDict(),
     ) -> Dict[str, jnp.ndarray]:
         rng, key = random.split(rng)
-        v, ldj = self.sample_forward(
-            key, tuple(state.params.values()), kwargs["n_samples"]
-        )
-        logp, aux = vmap(
-            lambda _params: fun(
-                unravel(_params),
+        v, ldj = self.sample_forward(key, state.params, kwargs["n_samples"])
+        neg_logp, aux = vmap(
+            lambda _rav_params: loss_fun(
+                self._unravel(_rav_params),
                 batch,
                 n_data=n_data,
                 mutable=state.mutable,
@@ -216,7 +218,7 @@ class NormalizingFlowTrainer(PosteriorTrainerABC):
                 calib_mutable=state.calib_mutable,
             )
         )(v)
-        loss = -(jnp.mean(logp) + jnp.mean(ldj))
+        loss = jnp.mean(neg_logp) - jnp.mean(ldj)
         if metrics is not None:
             preds = self.predict_fn(aux["outputs"])
             val_metrics = vmap(lambda p: self.compute_metrics(p, batch[1], metrics))(
