@@ -1,10 +1,12 @@
 from typing import Optional
 
-from fortuna.prob_model.posterior.name_to_posterior_state import \
-    NameToPosteriorState
+from fortuna.prob_model.posterior.name_to_posterior_state import NameToPosteriorState
 from fortuna.prob_model.posterior.state import PosteriorState
 from fortuna.training.mixin import WithCheckpointingMixin
-from fortuna.typing import OptaxOptimizer, Path
+from fortuna.typing import (
+    OptaxOptimizer,
+    Path,
+)
 
 
 class WithPosteriorCheckpointingMixin(WithCheckpointingMixin):
@@ -20,5 +22,5 @@ class WithPosteriorCheckpointingMixin(WithCheckpointingMixin):
             restore_checkpoint_path,
             optimizer,
             prefix,
-            name_to_train_state=NameToPosteriorState,
+            name_to_train_state=name_to_train_state,
         )
