@@ -1,5 +1,4 @@
-from fortuna.prob_model.posterior.posterior_mixin import \
-    WithPosteriorCheckpointingMixin
+from fortuna.prob_model.posterior.posterior_mixin import WithPosteriorCheckpointingMixin
 from fortuna.training.trainer import TrainerABC
 
 
