@@ -1,20 +1,42 @@
-from typing import Any, Callable, Dict, Optional, Tuple
+from typing import (
+    Any,
+    Callable,
+    Dict,
+    List,
+    Optional,
+    Tuple,
+)
 
-import jax.numpy as jnp
 from flax.core import FrozenDict
+from jax import device_get
 from jax.flatten_util import ravel_pytree
+import jax.numpy as jnp
 from jax.tree_util import tree_map
 
 from fortuna.prob_model.posterior.map.map_trainer import MAPTrainer
 from fortuna.prob_model.posterior.swag.swag_state import SWAGState
-from fortuna.training.trainer import JittedMixin, MultiDeviceMixin
-from fortuna.typing import Array, Batch
+from fortuna.training.callback import Callback
+from fortuna.training.trainer import (
+    JittedMixin,
+    MultiDeviceMixin,
+)
+from fortuna.typing import (
+    Array,
+    Batch,
+    Path,
+)
+from fortuna.utils.strings import encode_tuple_of_lists_of_strings_to_numpy
 
 
 class SWAGTrainer(MAPTrainer):
-    _mean_rav_params = None
-    _mean_squared_rav_params = None
-    _deviation_rav_params = None
+    def __init__(self, *, which_params: Optional[Tuple[List[str]]], **kwargs):
+        super(SWAGTrainer, self).__init__(**kwargs)
+        self._mean_rav_params = None
+        self._mean_squared_rav_params = None
+        self._deviation_rav_params = None
+        self._encoded_which_params = encode_tuple_of_lists_of_strings_to_numpy(
+            which_params
+        )
 
     def _update_state_with_stats(self, state: SWAGState) -> SWAGState:
         var = self._mean_squared_rav_params - self._mean_rav_params**2
@@ -28,6 +50,7 @@ class SWAGTrainer(MAPTrainer):
                 dev=self._deviation_rav_params
                 if not self.multi_device
                 else self._deviation_rav_params[None],
+                _encoded_which_params=self._encoded_which_params,
             )
         )
 
@@ -38,8 +61,9 @@ class SWAGTrainer(MAPTrainer):
         aux: Dict[str, Any],
         batch: Batch,
         metrics: Optional[Tuple[Callable[[jnp.ndarray, Array], float], ...]] = None,
+        callbacks: Optional[List[Callback]] = None,
         kwargs: FrozenDict[str, Any] = FrozenDict(),
-    ) -> Dict[str, jnp.ndarray]:
+    ) -> Tuple[SWAGState, Dict[str, jnp.ndarray]]:
         if "rank" not in kwargs:
             raise AttributeError(
                 """`rank` must be available in `kwargs` during training."""
@@ -68,16 +92,35 @@ class SWAGTrainer(MAPTrainer):
                 ),
                 axis=1,
             )
-        if (
-            self.save_checkpoint_dir
-            and self.save_every_n_steps is not None
-            and current_epoch % self.save_every_n_steps
-        ):
-            state = self._update_state_with_stats(state)
         return super().training_step_end(
-            current_epoch, state, aux, batch, metrics, kwargs
+            current_epoch, state, aux, batch, metrics, callbacks, kwargs
         )
 
+    def save_checkpoint(
+        self,
+        state: SWAGState,
+        save_checkpoint_dir: Path,
+        keep: int = 1,
+        force_save: bool = False,
+        prefix: str = "checkpoint_",
+    ) -> None:
+        state = self._update_state_with_stats(state)
+        super().save_checkpoint(state, save_checkpoint_dir, keep, force_save, prefix)
+
+    def on_train_end(self, state: SWAGState) -> SWAGState:
+        self.save_checkpoint(
+            state,
+            save_checkpoint_dir=self.save_checkpoint_dir,
+            keep=self.keep_top_n_checkpoints,
+            force_save=True,
+        )
+
+        if self.freeze_fun is not None:
+            state = state.replace(
+                params=self._get_all_params(state), frozen_params=None
+            )
+        return self._update_state_with_stats(state)
+
 
 class SWAGJittedMixin(JittedMixin):
     def on_train_end(self, state: SWAGState) -> SWAGState:
@@ -96,4 +139,12 @@ class JittedSWAGTrainer(SWAGJittedMixin, SWAGTrainer):
 
 
 class MultiDeviceSWAGTrainer(SWAGMultiDeviceMixin, SWAGTrainer):
-    pass
+    def on_train_end(self, state: SWAGState) -> SWAGState:
+        self.save_checkpoint(
+            state,
+            save_checkpoint_dir=self.save_checkpoint_dir,
+            keep=self.keep_top_n_checkpoints,
+            force_save=True,
+        )
+        state = device_get(tree_map(lambda x: x[0], state))
+        return self._update_state_with_stats(state)
