@@ -1,19 +1,38 @@
 import abc
-from typing import Callable, Dict, List, Optional, Tuple, Union
+from typing import (
+    Callable,
+    Dict,
+    List,
+    Optional,
+    Tuple,
+    Union,
+)
 
-import jax
+from flax.training.common_utils import shard_prng_key
+from jax import (
+    jit,
+    lax,
+    pmap,
+    random,
+)
+from jax._src.prng import PRNGKeyArray
 import jax.numpy as jnp
 import jax.scipy as jsp
-from jax import jit, lax, pmap, random
-from jax._src.prng import PRNGKeyArray
 from jax.tree_util import tree_map
 
-from fortuna.data.loader import (DataLoader,
-                                 DeviceDimensionAugmentedDataLoader,
-                                 DeviceDimensionAugmentedInputsLoader,
-                                 InputsLoader, TargetsLoader)
+from fortuna.data.loader import (
+    DataLoader,
+    DeviceDimensionAugmentedLoader,
+    InputsLoader,
+    TargetsLoader,
+)
 from fortuna.prob_model.posterior.base import Posterior
-from fortuna.typing import Array, Batch, CalibMutable, CalibParams
+from fortuna.typing import (
+    Array,
+    Batch,
+    CalibMutable,
+    CalibParams,
+)
 from fortuna.utils.random import WithRNG
 
 
@@ -30,13 +49,92 @@ class Predictive(WithRNG):
         self.likelihood = posterior.joint.likelihood
         self.posterior = posterior
 
+    ####
+    def ensemble_log_prob(
+        self,
+        data_loader: DataLoader,
+        n_posterior_samples: int = 30,
+        rng: Optional[PRNGKeyArray] = None,
+        distribute: bool = True,
+        **kwargs,
+    ) -> jnp.ndarray:
+        r"""
+        Compute the log-likelihood at each posterior sample, that is
+
+        .. math::
+            \log p(y|x, theta^{(i)}),
+
+        where:
+         - :math:`x` is an observed input variable;
+         - :math:`y` is an observed target variable;
+         - :math:`theta^{(i)}` is a sample from the posterior.
+
+        Parameters
+        ----------
+        data_loader : DataLoader
+            A data loader.
+        n_posterior_samples : int
+            Number of posterior samples to draw in order to compute the log -ikelihood.
+            that would be produced using the posterior distribution state.
+        rng : Optional[PRNGKeyArray]
+            A random number generator. If not passed, this will be taken from the attributes of this class.
+        distribute: bool
+            Whether to distribute computation over multiple devices, if available.
+
+        Returns
+        -------
+        jnp.ndarray
+            An array of log-likelihood values at each posterior sample for each data point.
+        """
+
+        if rng is None:
+            rng = self.rng.get()
+
+        log_probs = self._loop_fun_through_data_loader(
+            self._batched_ensemble_log_prob,
+            data_loader,
+            n_posterior_samples,
+            rng,
+            distribute,
+            **kwargs,
+        )
+
+        return log_probs.swapaxes(0, 1)
+
+    def _batched_ensemble_log_prob(
+        self,
+        batch: Batch,
+        n_posterior_samples: int = 30,
+        rng: Optional[PRNGKeyArray] = None,
+        **kwargs,
+    ) -> Union[jnp.ndarray, Tuple[jnp.ndarray, dict]]:
+        if rng is None:
+            rng = self.rng.get()
+        keys = random.split(rng, n_posterior_samples)
+
+        def _lik_log_batched_prob(key):
+            sample = self.posterior.sample(inputs=batch[0], rng=key)
+            return self.likelihood._batched_log_prob(
+                sample.params,
+                batch,
+                mutable=sample.mutable,
+                calib_params=sample.calib_params,
+                calib_mutable=sample.calib_mutable,
+                **kwargs,
+            )
+
+        return lax.map(_lik_log_batched_prob, keys).swapaxes(
+            0, 1
+        )  # notice that we use lax.map and not vmap to avoid parallel posterior sampling
+
+    ####
     def log_prob(
         self,
         data_loader: DataLoader,
         n_posterior_samples: int = 30,
         rng: Optional[PRNGKeyArray] = None,
         distribute: bool = True,
-        **kwargs
+        **kwargs,
     ) -> jnp.ndarray:
         r"""
         Estimate the predictive log-probability density function (a.k.a. log-pdf), that is
@@ -75,7 +173,7 @@ class Predictive(WithRNG):
             n_posterior_samples,
             rng,
             distribute,
-            **kwargs
+            **kwargs,
         )
 
     def _batched_log_prob(
@@ -83,7 +181,7 @@ class Predictive(WithRNG):
         batch: Batch,
         n_posterior_samples: int = 30,
         rng: Optional[PRNGKeyArray] = None,
-        **kwargs
+        **kwargs,
     ) -> Union[jnp.ndarray, Tuple[jnp.ndarray, dict]]:
         if rng is None:
             rng = self.rng.get()
@@ -97,7 +195,7 @@ class Predictive(WithRNG):
                 mutable=sample.mutable,
                 calib_params=sample.calib_params,
                 calib_mutable=sample.calib_mutable,
-                **kwargs
+                **kwargs,
             )
 
         return jsp.special.logsumexp(
@@ -114,7 +212,7 @@ class Predictive(WithRNG):
         calib_params: Optional[CalibParams] = None,
         calib_mutable: Optional[CalibMutable] = None,
         rng: Optional[PRNGKeyArray] = None,
-        **kwargs
+        **kwargs,
     ) -> Union[jnp.ndarray, Tuple[jnp.ndarray, Dict]]:
         if return_aux is None:
             return_aux = []
@@ -122,7 +220,7 @@ class Predictive(WithRNG):
         unsupported_aux = [s for s in return_aux if s not in supported_aux]
         if sum(unsupported_aux) > 0:
             raise Exception(
-                """The auxiliary objects {} are unknown. Please make sure that all elements of `return_aux` 
+                """The auxiliary objects {} are unknown. Please make sure that all elements of `return_aux`
                             belong to the following list: {}""".format(
                     unsupported_aux, supported_aux
                 )
@@ -147,7 +245,7 @@ class Predictive(WithRNG):
                     calib_mutable=calib_mutable
                     if calib_mutable is not None
                     else sample.calib_mutable,
-                    **kwargs
+                    **kwargs,
                 )
 
             outs = lax.map(_lik_log_joint_prob, keys)
@@ -167,7 +265,7 @@ class Predictive(WithRNG):
                     return_aux=return_aux,
                     calib_params=calib_params,
                     calib_mutable=calib_mutable,
-                    **kwargs
+                    **kwargs,
                 )
 
             outs = lax.map(_lik_log_joint_prob, ensemble_outputs)
@@ -184,6 +282,35 @@ class Predictive(WithRNG):
             return log_pred, aux
         return log_pred
 
+    def _batched_negative_log_joint_prob(
+        self,
+        batch: Batch,
+        n_data: int,
+        n_posterior_samples: int = 30,
+        return_aux: Optional[List[str]] = None,
+        ensemble_outputs: Optional[jnp.ndarray] = None,
+        calib_params: Optional[CalibParams] = None,
+        calib_mutable: Optional[CalibMutable] = None,
+        rng: Optional[PRNGKeyArray] = None,
+        **kwargs,
+    ) -> Union[jnp.ndarray, Tuple[jnp.ndarray, Dict]]:
+        outs = self._batched_log_joint_prob(
+            batch,
+            n_data,
+            n_posterior_samples,
+            return_aux,
+            ensemble_outputs,
+            calib_params,
+            calib_mutable,
+            rng,
+            **kwargs,
+        )
+        if len(return_aux) > 0:
+            loss, aux = outs
+            loss *= -1
+            return loss, aux
+        return -outs
+
     def sample(
         self,
         inputs_loader: InputsLoader,
@@ -191,7 +318,7 @@ class Predictive(WithRNG):
         return_aux: Optional[List[str]] = None,
         rng: Optional[PRNGKeyArray] = None,
         distribute: bool = True,
-        **kwargs
+        **kwargs,
     ) -> Union[jnp.ndarray, Tuple[jnp.ndarray, Dict[str, jnp.ndarray]]]:
         r"""
         Sample from an approximation of the predictive distribution for each input data point, that is
@@ -225,16 +352,13 @@ class Predictive(WithRNG):
         if not rng:
             rng = self.rng.get()
 
-        if distribute and jax.local_device_count() <= 1:
-            distribute = False
-
         def fun(_inputs):
             return self._batched_sample(
                 _inputs, n_target_samples, return_aux, rng, **kwargs
             )
 
         if distribute:
-            inputs_loader = DeviceDimensionAugmentedInputsLoader(inputs_loader)
+            inputs_loader = DeviceDimensionAugmentedLoader(inputs_loader)
             fun = pmap(fun)
             if return_aux is None or len(return_aux) == 0:
                 return jnp.concatenate(
@@ -242,7 +366,7 @@ class Predictive(WithRNG):
                         self.likelihood._unshard_array(fun(inputs))
                         for inputs in inputs_loader
                     ],
-                    0,
+                    1,
                 )
             else:
                 samples, aux_outputs = [], []
@@ -281,7 +405,7 @@ class Predictive(WithRNG):
         n_target_samples: int = 1,
         return_aux: Optional[List[str]] = None,
         rng: Optional[PRNGKeyArray] = None,
-        **kwargs
+        **kwargs,
     ) -> jnp.ndarray:
         if return_aux is None:
             return_aux = []
@@ -302,7 +426,7 @@ class Predictive(WithRNG):
                 calib_mutable=_post_sample.calib_mutable,
                 return_aux=return_aux,
                 rng=key2,
-                **kwargs
+                **kwargs,
             )
             if len(return_aux) > 0:
                 _samples, aux = outs
@@ -418,27 +542,29 @@ class Predictive(WithRNG):
             rng = self.rng.get()
         keys = random.split(rng, n_output_samples)
 
-        if distribute and jax.local_device_count() <= 1:
-            distribute = False
-
         if distribute:
-            inputs_loader = DeviceDimensionAugmentedInputsLoader(inputs_loader)
+            inputs_loader = DeviceDimensionAugmentedLoader(inputs_loader)
 
-        @jit
         def _sample(key, _inputs):
             sample = self.posterior.sample(inputs=_inputs, rng=key)
             return self.likelihood.model_manager.apply(
                 params=sample.params, inputs=_inputs, mutable=sample.mutable
             )
 
+        _sample = pmap(_sample) if distribute else jit(_sample)
+
         iterable = []
         size = 0
         for inputs in inputs_loader:
-            size += inputs.shape[0]
+            size += (
+                inputs.shape[0]
+                if not isinstance(inputs, dict)
+                else inputs[list(inputs.keys())[0]].shape[0]
+            )
             if distribute:
-                outputs = pmap(
-                    lambda _inputs: lax.map(lambda key: _sample(key, _inputs), keys)
-                )(inputs)
+                outputs = jnp.stack(
+                    list(map(lambda key: _sample(shard_prng_key(key), inputs), keys))
+                )
                 outputs = self._unshard_ensemble_arrays(outputs)
             else:
                 outputs = lax.map(lambda key: _sample(key, inputs), keys)
@@ -825,16 +951,13 @@ class Predictive(WithRNG):
         n_posterior_samples: int,
         rng: PRNGKeyArray,
         distribute: bool = True,
-        **kwargs
+        **kwargs,
     ) -> Array:
-        if distribute and jax.local_device_count() <= 1:
-            distribute = False
-
         def fun2(_inputs):
             return fun(_inputs, n_posterior_samples, rng, **kwargs)
 
         if distribute:
-            inputs_loader = DeviceDimensionAugmentedInputsLoader(inputs_loader)
+            inputs_loader = DeviceDimensionAugmentedLoader(inputs_loader)
             fun2 = pmap(fun2)
             return jnp.concatenate(
                 [
@@ -853,16 +976,13 @@ class Predictive(WithRNG):
         n_posterior_samples: int,
         rng: PRNGKeyArray,
         distribute: bool = True,
-        **kwargs
+        **kwargs,
     ) -> Array:
-        if distribute and jax.local_device_count() <= 1:
-            distribute = False
-
         def fun2(_batch):
             return fun(_batch, n_posterior_samples, rng, **kwargs)
 
         if distribute:
-            data_loader = DeviceDimensionAugmentedDataLoader(data_loader)
+            data_loader = DeviceDimensionAugmentedLoader(data_loader)
             fun2 = pmap(fun2)
             return jnp.concatenate(
                 [self.likelihood._unshard_array(fun2(batch)) for batch in data_loader],
@@ -878,20 +998,17 @@ class Predictive(WithRNG):
         n_posterior_samples: int,
         rng: PRNGKeyArray,
         distribute: bool = True,
-        **kwargs
+        **kwargs,
     ) -> Array:
-        if distribute and jax.local_device_count() <= 1:
-            distribute = False
-
         def fun2(_inputs):
             return fun(_inputs, n_posterior_samples, rng, **kwargs)
 
         if distribute:
-            inputs_loader = DeviceDimensionAugmentedInputsLoader(inputs_loader)
+            inputs_loader = DeviceDimensionAugmentedLoader(inputs_loader)
             fun2 = pmap(fun2)
             return jnp.concatenate(
                 [
-                    self.likelihood._unshard_array(fun2(inputs))
+                    self._unshard_ensemble_arrays(fun2(inputs))
                     for inputs in inputs_loader
                 ],
                 1,
